import argparse
import math

from PIL import Image

from stego.utility import *

def stegd(img, sec_key):
    """Extract the hidden information, image and/or text, from an image
    :param img: Cover image
    :param sec_key: Secret key, usually is a word or sentence
    """

    try:
        # Open the Stego image
        with Image.open(img) as cover_img:
            cov_pxs = list(cover_img.getdata())
    except:
        raise IOError('Can\'t open image!')

    if not sec_key:
        raise ValueError('Secure Key is needed!')

    #### Get size of hidden information
    l = math.ceil(math.log(len(cov_pxs)//8, 2))

    hm_wid = hm_len = ht_len = 0
    for px in cov_pxs[:l]:
        # Get size of hidden image
        hm_wid = 2 * hm_wid + (px[1] & 1)
        hm_len = 2 * hm_len + (px[2] & 1)

        # Get size of hidden text
        ht_len = 2 * ht_len + (px[0] & 1)

    # Get secret key bit stream
    sec_bs = str2bs(sec_key)
    len_sec_bs = len(sec_bs)

    #### Start to extract hidden information bit by bit
    cnt, i, ret = 0, 0, []
    
    len_lg_bs = max(hm_wid * hm_len * 3 * 8, ht_len * 8)
    len_sh_bs = min(hm_wid * hm_len * 3 * 8, ht_len * 8)

    sk_bs = [int(k) for k in str2bs(sec_key)]
    len_sk_bs = len(sk_bs)

    lg_bs, sh_bs = [], []
    for px in cov_pxs[l:]:

        if px[0] ^ sk_bs[i% len_sk_bs]:
            lg_bs.append(px[1] & 1)
            if i < len_sh_bs:
                sh_bs.append(px[2] & 1)

        else:
            lg_bs.append(px[2] & 1)
            if i < len_sh_bs:
                sh_bs.append(px[1] & 1)
        i += 1
        if i >= len_lg_bs:
            break

    lg_bs = ''.join([str(c) for c in lg_bs])
    sh_bs = ''.join([str(c) for c in sh_bs])

    if hm_wid * hm_len * 3 > ht_len:
        hid_img, hid_text = bs2img(lg_bs), bs2str(sh_bs)
    else:
        hid_img, hid_text = bs2img(sh_bs), bs2str(lg_bs)

    if len(hid_img) > 0:
        img = Image.new('RGB', (hm_wid, hm_len))
        img.putdata(hid_img)
        img.save('hidden_img.png')

    if len(hid_text) > 0:
        with open('hidden_text.txt', 'w+') as ht:
            ht.write(hid_text)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', help='Path of cover image')
    parser.add_argument('-s', help='Secure key')

    args = parser.parse_args()
    stegd(args.c, args.s)

if __name__ == '__main__':
    main()


def str2bs(s):
    """
    Convert string to 8 bits binary bit stream
    :param s: A string
    :return: 1D bit stream
    """
    result = [bin(ord(ch))[2:].zfill(8) for ch in s]
    return ''.join(result)

def replace_lsb(i, b):
    """
    Change the last bit of integer i to b
    :param i: An integer number
    :param b: Bit number 1 or 0
    :return: Integer number after changing last bit of i
    """
    result = i | int(b) if int(b) == 1 else i & 0xfe
    return result

def bs2img(s):
    """
    Get RGB values from input bit stream
    :param s: Bit stream
    :return: RGB tuples
    """
    if len(s) % 24 != 0:
        print('Invalid bit stream!')
        raise StopIteration

    vals = [int(s[i:i+8], 2) for i in range(0, len(s), 8)]
    pxs = [(vals[i], vals[i+1], vals[i+2]) for i in range(0, len(vals), 3)]
    return pxs

def bs2str(s):
    """
    Get text string from input bit stream
    :param s: Bit stream of a string
    :return: String
    """
    if len(s) % 8:
        print('Invalid bit stream!')
        raise ValueError

    result = [chr(int(s[i:i+8], 2)) for i in range(0, len(s), 8)]
    return ''.join(result)


import argparse
import math

from PIL import Image

from stego.utility import *

def stege(fp_cover_img, fp_hid_img, fp_hid_text, sec_key):
    """Encrypt image and/or text into cover image serially. With this method, hidden image and hidden text bit streams are concatenated before being encrypted.
    :param fp_cover_img: Path of cover image
    :param fp_hid_img: Path of hidden image
    :param fp_hid_text: Value of hidden text
    :param sec_key: Value of secret key
    """

    if not sec_key:
        raise ValueError('Secure Key is needed!')

    if not fp_hid_img and not fp_hid_text:
        raise ValueError('Nothing to hide!')

    try:
        cimg = Image.open(fp_cover_img)
        cov_pxs = list(cimg.getdata())
    except:
        raise IOError('Can\'t open cover image!')

    # hidden text and image are initialized to empty strings
    hid_img_width, hid_img_height = 0, 0
    hid_pxs = []
    hid_text, ht_bs, hm_bs = '', '', ''

    # encode the size of hidden image and text
    l = math.ceil(math.log(len(cov_pxs)/8, 2))

    if fp_hid_img:
        try:
            with Image.open(fp_hid_img) as himg:
                # get pixels of images
                hid_pxs = [px[:3] for px in himg.getdata()]
                hid_img_width = himg.size[0]
                hid_img_height = himg.size[1]

        except:
            raise IOError('Can\'t open hidden image!')
        else:
            # the number of pixels in hidden image cannot exceed one third of the number of pixels in cover image
            if len(hid_pxs) * 3 * 8 > len(cov_pxs) - l:
                raise ValueError('Hidden image is too big!')
            hm_bs = ''.join([bin(val)[2:].zfill(8) for px in hid_pxs for val in px])

    if fp_hid_text:
        try:
            with open(fp_hid_text) as ht_file:
                hid_text = ht_file.read().strip()
        except:
            raise IOError('Can\'t open hidden text file!')
        else:
            # hidden text cannot be longer than the number of pixels in cover image
            if len(hid_text) * 8 > len(cov_pxs) - l:
                raise ValueError('Hidden text is too long!')
            ht_bs = str2bs(hid_text)

    hm_width_bs = bin(hid_img_width)[2:].zfill(l)
    hm_height_bs = bin(hid_img_height)[2:].zfill(l)
    ht_len_bs = bin(len(hid_text))[2:].zfill(l)

    # ret stores the pixels of encoded cover image
    ret = []
    i = 0

    # hidden image and text sizes are encoded in the first l pixels
    # text size is hidden in red, image size is hidden in blue and green
    for px in cov_pxs[:l]:
        ret.append((replace_lsb(px[0], ht_len_bs[i]), replace_lsb(px[1], hm_width_bs[i]), replace_lsb(px[2], hm_height_bs[i])))
        i += 1

    # encode the content
    hid_lg_bs, hid_sh_bs = hm_bs, ht_bs
    if len(hid_text) > len(hid_pxs) * 3:
        hid_lg_bs, hid_sh_bs = ht_bs, hm_bs

    len_lg_bs = len(hid_lg_bs)
    len_hm_bs = len(hid_sh_bs)

    # convert secure key to binary bit stream
    sk_bs = [int(k) for k in str2bs(sec_key)]
    len_sk_bs, i = len(sk_bs), 0

    for px in cov_pxs[l:]:
        new_gpx, new_bpx = px[1], px[2]
        if px[0] ^ sk_bs[i % len_sk_bs]:
            new_gpx = replace_lsb(px[1], int(hid_lg_bs[i]))
            if i < len(hid_sh_bs):
                new_bpx = replace_lsb(px[2], int(hid_sh_bs[i]))

        else:
            new_bpx = replace_lsb(px[2], int(hid_lg_bs[i]))
            if i < len(hid_sh_bs):
                new_gpx = replace_lsb(px[1], int(hid_sh_bs[i]))
        ret.append((px[0], new_gpx, new_bpx))
        i += 1
        if i >= len_lg_bs:
            break

    # save encoded image
    img = Image.new('RGB', cimg.size)
    img.putdata(ret + cov_pxs[i+l:])
    img.save('-'.join([fp_cover_img.split('.')[0], 'encoded.png']))

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', help='Path of cover image')
    parser.add_argument('-s', help='Secure key')
    parser.add_argument('-hi', help='Path of hidden image')
    parser.add_argument('-ht', help='Path of hidden text file')

    args = parser.parse_args()
    try:
        stege(args.c, args.hi, args.ht, args.s)
    except Exception as e:
        print(str(e))

if __name__ == '__main__':
    main()


from setuptools import setup

with open('README.md') as readme:
    description = readme.read()

setup(
    name='Stego',
    version='0.0.1',
    description='Hide your secret',
    author='Linqiang Ouyang (Jimmy)',
    author_email='jimmyou587@gmail.com',
    packages=['stego'],
    license='MIT',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    # What does your project relate to?
    keywords='python, steganography',
    install_requires=['pillow'],
    scripts=['bin/stege', 'bin/stegd'],
    test_suite='nose.collector',
    tests_require=['nose']
    )